#!/usr/bin/perl
# UseModWiki version 0.82 (July 15, 2000)
# Copyright (C) 2000 Clifford A. Adams
#    <caadams@frontiernet.net> or <usemod@usemod.com>
# Based on the GPLed AtisWiki 0.3  (C) 1998 Markus Denker
#    <marcus@ira.uka.de>
# ...which was based on
#    the LGPLed CVWiki CVS-patches (C) 1997 Peter Merel
#    and The Original WikiWikiWeb  (C) Ward Cunningham
#       <ward@c2.com> (code reused with permission)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
#    Free Software Foundation, Inc.
#    59 Temple Place, Suite 330
#    Boston, MA 02111-1307 USA

$| = 1;    # Do not buffer output
undef $/;  # Read complete files
use strict;
use vars qw(%page %FullPage %InterSite %SaveUrl
  @HtmlStack @RcDays
  $LockDir $DataDir $HtmlDir $InterFile $RcFile $FullUrl
  $SiteName $HomePage $ScriptTZ $LogoUrl $RcDefault $IndentLimit
  $RecentTop $EditAllowed $UseDiff $UseSubpage $UseCache $RawHtml
  $SimpleLinks $UseUmlauts $LogoLeft $ShowEdits
  $LinkPattern $UpperLetter $LowerLetter $AnyLetter $LpA $LpB
  $InterLinkPattern $InterSitePattern $InterSiteInit
  $UrlProtocols $UrlPattern $ImageExtensions $ImagePattern
  $PageHtml $SaveHtmlStack $SaveUrlIndex
  $MainPage $OpenId $q $FS $FS1 $FS2 $FS3 $QDelim
  $ScriptName $BrowseCode $OtherCode);

# Configuration:
# Note: The script can auto-create the Data/Lock/Hist/Html directories
$DataDir     = "/tmp/mywikidb";     # Main page database
$LockDir     = "$DataDir/lock";     # DB is locked if this exists
$HtmlDir     = "$DataDir/html";     # Stores HTML versions
$InterFile   = "$DataDir/intermap"; # Interwiki site->url map
$RcFile      = "$DataDir/rclog";    # New RecentChanges logfile

$FullUrl     = "";  # Set to full URL to use redirect, for example:
# $FullUrl   = "http://www.usemod.com/cgi-bin/mywiki.pl";

$SiteName    = "Wiki";          # Name of site (used for titles)
$HomePage    = "HomePage";      # Home page (must be valid LinkPattern)
$LogoUrl     = "/wiki.gif";     # URL for site logo ("" for no logo)
$ENV{PATH}   = "/usr/bin/";     # Path used to find "diff"
$ScriptTZ    = "";              # Local time zone ("" means do not print)
$RcDefault   = 30;              # Default number of RecentChanges days
@RcDays      = qw(1 3 7 30 90); # Days for links on RecentChanges
$IndentLimit = 20;              # Maximum depth of nested lists

$UseSubpage  = 1;       # 1 = use subpages,       0 = do not use subpages
$UseCache    = 0;       # 1 = cache HTML pages,   0 = generate every page
$EditAllowed = 1;       # 1 = editing allowed,    0 = read-only
$LogoLeft    = 0;       # 1 = logo on left,       0 = logo on right
$RecentTop   = 1;       # 1 = recent on top,      0 = recent on bottom
$UseDiff     = 1;       # 1 = use diff features,  0 = do not use diff
$ShowEdits   = 0;       # 1 = show minor edits,   0 = hide edits by default
$RawHtml     = 0;       # 1 = allow <HTML> tag,   0 = no raw HTML in pages
$SimpleLinks = 0;       # 1 = only letters,       0 = allow _ and numbers
$UseUmlauts  = 0;       # 1 = allow link umlauts, 0 = only A-Za-z chars

# LinkPattern definitions
$UpperLetter = "[A-Z";
$LowerLetter = "[a-z";
$AnyLetter   = "[A-Za-z";

# Very incomplete list of non-English characters
if ($UseUmlauts) {
  my ($upper, $lower);
  $upper = "\xc4\xd6\xdc";
  $lower = "\xdf\xe4\xf6\xfc";
  $UpperLetter .= $upper;
  $LowerLetter .= $lower;
  $AnyLetter   .= $upper . $lower;
}
if (!$SimpleLinks) {
  $AnyLetter .= "_0-9";
}
$UpperLetter .= "]"; $LowerLetter .= "]"; $AnyLetter .= "]";

# Main link pattern: lowercase between uppercase, then anything
$LpA = $UpperLetter . "+" . $LowerLetter . "+" . $UpperLetter
       . $AnyLetter . "*";
# Optional subpage link pattern: uppercase, lowercase, then anything
$LpB = $UpperLetter . "+" . $LowerLetter . "+" . $AnyLetter . "*";

if ($UseSubpage) {
  # Loose pattern: If subpage is used, subpage may be simple name
  $LinkPattern = "((($LpA)?\\/$LpB)|$LpA)";
  # Strict pattern: both sides must be the main LinkPattern
  # $LinkPattern = "((($LpA)?\\/)?$LpA)";
} else {
  $LinkPattern = "($LpA)";
}
$QDelim = '("")?';     # Optional quote delimiter (not in output)
$LinkPattern .= $QDelim;

# Url-style links are delimited by one of:
#   1.  Whitespace (kept in output)
#   2.  Left angle-bracket (<)  (kept in output)
#   3.  A single double-quote (")  (kept in output)
#   4.  A double double-quote ("") (removed from output)

# Inter-site convention: sites must start with uppercase letter
# (Uppercase letter avoids confusion with URLs)
$InterSitePattern = $UpperLetter . $AnyLetter . "+";
$InterLinkPattern = "($InterSitePattern:[^\\s\"<]+)" . $QDelim;
$UrlProtocols = "(http|ftp|afs|news|nntp|mid|cid|mailto|wais|"
                . "prospero|telnet|gopher)";
$UrlPattern = "(($UrlProtocols):[^\\s\"<]+)" . $QDelim;
$ImageExtensions = "(gif|jpg|png|bmp|jpeg)";
$ImagePattern = "(http:[^\\s\"<]+\\.$ImageExtensions)" . $QDelim;

$BrowseCode = ""; # Comment next line to always compile (slower)
#$BrowseCode = <<'#END_OF_BROWSE_CODE';
use CGI;
use CGI::Carp qw(fatalsToBrowser);
$q = new CGI;

sub DoBrowseRequest {
  my ($id, $action, $text, $isDiff);

  &InitRequest() or return 1;
  if (!$q->param) {            # No parameter
    &BrowsePage($HomePage, 0, 0);
    return 1;
  }
  $id = &GetParam("keywords", "");
  if ($id) {                   # Just ScriptName?PageName
    &BrowsePage($id, 0, 0) if &ValidId($id);
    return 1;
  }
  $action = lc(&GetParam("action", ""));
  $isDiff = &GetParam("diff", 0);
  $id = &GetParam("id", "");
  if ($action eq "browse") {
    &BrowsePage($id, $isDiff, 0) if &ValidId($id);
    return 1;
  } elsif ($action eq "rebrowse") {
    &BrowsePage($id, $isDiff, 1) if &ValidId($id);
    return 1;
  } elsif ($action eq "rc") {
    &BrowsePage("RecentChanges", 0, 0);
    return 1;
  }
  return 0;  # Request not handled
}

sub InitRequest {
  my @ScriptPath = split('/', "$ENV{SCRIPT_NAME}");

  $ScriptName = pop(@ScriptPath);  # Name used in links
  $FS = "\263"; $FS1 = $FS . "1"; $FS2 = $FS . "2"; $FS3 = $FS . "3";
  $MainPage = ".";  # Directory of the main page (used for subpages)
  $InterSiteInit = 0;
  %InterSite = ();
  $OpenId = "";
  if (!chdir $DataDir) {
    mkdir($DataDir,0775)  if (!(-d $DataDir));
    if (!chdir $DataDir) {
      &ReportError("Could not go to or create $DataDir: $!");
      return 0;
    }
  }
  return 1;
}

sub GetParam {
  my ($name, $default) = @_;
  my $result;

  $result = $q->param($name);
  if (!defined($result)) {
    $result = $default;
  }
  return $result;
}

sub BrowsePage {
  my ($id, $showDiff, $reBrowse) = @_;
  my ($fullHtml, $diffHtml);

  %page = &GetCurrentPage($id);
  # Handle a single-level redirect
  if ((!$reBrowse) && (substr($page{text}, 0, 10) eq "#REDIRECT ")) {
    ($id) = ($page{text} =~ /\#REDIRECT\s+(\S+)/);
    return  if (!&ValidId($id));
    &ReBrowsePage($id, 1);
    return;
  }
  $MainPage = $id;
  $MainPage =~ s|/.*||;  # Only the main page name (remove subpage)

  $fullHtml = &GetHeader($id, &QuoteHtml($id), $reBrowse);
  if ($showDiff) {
    $fullHtml .= "<h2>Difference (from prior version)</h2>\n"
                 . &DiffToHTML($page{prevdiff}) . "<hr>\n";
  }
  $fullHtml .= &WikiToHTML($page{text}) . "<hr>\n";
  if ($id eq "RecentChanges") {
    print $fullHtml;
    &ActionRc();
    print "<hr>\n", &GetFooterText($id);
    return;
  }
  $fullHtml .= &GetFooterText($id);
  print $fullHtml;
  return  if ($showDiff);  # No cache for diff version
  &UpdateHtmlCache($id, $fullHtml)  if $UseCache;
}

# Consider handling diff links?  (needed?)
sub ReBrowsePage {
  my ($id, $reBrowse) = @_;

  if ($FullUrl ne "") {   # Location: style redirect is set up
    if ($reBrowse) {      # Target of #REDIRECT (loop breaking)
      $id = "action=rebrowse&id=$id";
    }
    print &GetRedirectPage($id);
  } else {
    &BrowsePage($id, 0, $reBrowse);
  }
}

sub DiffToHTML {
  my ($html) = @_;

  $html =~ s/\n--+//g;
  # Note: Need spaces before <br> to be different from diff section.
  $html =~ s/(^|\n)(\d+.*c.*)/$1 <br><strong>Changed: $2<\/strong><br>/g;
  $html =~ s/(^|\n)(\d+.*d.*)/$1 <br><strong>Removed: $2<\/strong><br>/g;
  $html =~ s/(^|\n)(\d+.*a.*)/$1 <br><strong>Added: $2<\/strong><br>/g;
  $html =~ s/\n((<.*\n)+)/&ColorDiff($1,"ffffaf")/ge;
  $html =~ s/\n((>.*\n)+)/&ColorDiff($1,"cfffcf")/ge;
  return $html;
}

sub ColorDiff {
  my ($diff, $color) = @_;

  $diff =~ s/(^|\n)[<>]/$1/g;
  $diff = &QuoteHtml($diff);
  # Do some of the Wiki markup rules:
  %SaveUrl = ();
  $SaveUrlIndex = 0;
  $diff =~ s/$FS//g;
  $diff =~ s/('*)'''(.*?)'''/$1<strong>$2<\/strong>/g;
  $diff =~ s/''(.*?)''/<em>$1<\/em>/g;
  $diff =~ s/\&lt;b\&gt;(.*?)\&lt;\/b\&gt;/<b>$1<\/b>/gi;
  $diff =~ s/\&lt;i\&gt;(.*?)\&lt;\/i\&gt;/<i>$1<\/i>/gi;
  $diff =~ s/----+/<hr>/g;
  $diff =~ s/$UrlPattern/&StoreUrl($1)/ge;
  $diff =~ s/$InterLinkPattern/&InterPageLink($1)/geo;
  $diff =~ s/$LinkPattern/&GetPageOrEditLink($1)/geo;
  $diff =~ s/\r?\n/<br>/g;
  $diff =~ s/$FS(\d+)$FS/$SaveUrl{$1}/ge;

  return "<table width=\"95\%\" bgcolor=#$color><tr><td>\n" . $diff
         . "</td></tr></table>\n";
}

sub QuoteHtml {
  my ($html) = @_;

  $html =~ s/&/&amp;/g;
  $html =~ s/</&lt;/g;
  $html =~ s/>/&gt;/g;
  return $html;
}

sub InterPageLink {
  my ($id) = @_;
  my ($name, $site, $remotePage, $url);

  $name = $id;
  ($site, $remotePage) = split(/:/, $id,2);
  $url = &GetSiteUrl($site);
  return $id  if ($url eq "");
  $remotePage =~ s/&amp;/&/g;  # Unquote common URL HTML
  $url .= $remotePage;
  $SaveUrl{$SaveUrlIndex} = "<a href=\"$url\">$name</a>";
  return $FS . $SaveUrlIndex++ . $FS;
}

sub GetSiteUrl {
  my ($site) = @_;
  my $data;
  my $url = "";

  if (!$InterSiteInit) {
    $InterSiteInit = 1;
    open(IN, "<$InterFile") or return "";  # Maybe later warn?
    $data=<IN>;
    close IN;
    %InterSite = split(/\s+/, $data);  # Later consider defensive code
  }
  $url = $InterSite{$site}  if (defined($InterSite{$site}));
  return $url;
}

sub GetPageLink {
  my ($id) = @_;
  my $name = $id;

  $id =~ s|^/|$MainPage/|;
  return "<a href=\"$ScriptName\?$id\">$name</a>";
}

sub GetEditLink {
  my ($id, $linkname) = @_;

  return "<a href=\"$ScriptName?action=Edit&id=$id\">$linkname</a>";
}

sub GetPageOrEditLink {
  my ($id) = @_;
  my $name = $id;

  $id =~ s|^/|$MainPage/|;
  if (-f (&GetPageDirectory($id) . "$id.db")) {  # Page exists
    return &GetPageLink($name);
  }
  return $name . &GetEditLink($id,"?");
}

sub GetSearchLink {
  my ($id) = @_;
  my $name = $id;

  $id =~ s|.+/|/|;  # Subpage match: search for just /SubName
  return "<a href=\"$ScriptName\?search=$id\">$name</a>\n";
}

sub GetHeader {
  my ($id, $title, $reBrowse) = @_;
  my $header = "";
  my $LogoImage = "";
  my $result = "";

  if ($LogoUrl ne "") {
    $LogoImage = "img src=\"$LogoUrl\" alt=\"Home\" border=0";
    if (!$LogoLeft) {
      $LogoImage .= " align=\"right\"";
    }
    $header = "<a href=\"$ScriptName\"><$LogoImage></a>";
  }
  $result = $q->header()
          . $q->start_html('-title' => "$SiteName: $title",
                           '-BGCOLOR' => 'white');
  if ($reBrowse) {
    $result .= $q->h3("(redirected from another page)");
  }
  if ($id ne "") {
    $result .= $q->h1($header . &GetSearchLink($id));
  } else {
    $result .= $q->h1($header . $title);
  }
  return $result;
}

sub GetFooterText {
  my ($id) = @_;
  my $last_edit = "";

  if ($page{timestamp} ne "") {
    $last_edit = " (last edited " . &TimeToText($page{timestamp});
    if ($UseDiff) {
      $last_edit .= " <a href=\"$ScriptName\?action=browse&diff=1&"
                    . "id=$id\">(diff)</a>";
    }
    $last_edit .= ")";
  }
  return $q->startform() . &GetGotoBar($id) .
    &GetEditLink($id, "Edit text of this page") . $last_edit . "<br>" .
    "Search:" .  $q->textfield(-name=>'search',-size=>20) .
    $q->endform . $q->end_html;
}

sub GetGotoBar {
  my ($id) = @_;

  my $main = $id;
  $main =~ s|/.*||;  # Only the main page name (remove subpage)
  my $bartext = "Goto: " . &GetPageLink($HomePage) . " | "
            . &GetPageLink("RecentChanges");
  if ($id =~ m|/|) {
    $bartext .= " | " . &GetPageLink($main);
  }
  $bartext .= "<br>\n";
  return $bartext;
}

sub StackHtml {
  my ($code, $depth) = @_;

  return  if ($depth < 1);  # Protect from bad depth
  if ($depth > $IndentLimit) {
    $depth = $IndentLimit;
  }
  $SaveHtmlStack = 1;
  while (@HtmlStack > $depth) {
    $PageHtml .=  "</" . pop(@HtmlStack) . ">\n";
  }
  while (@HtmlStack < $depth) {
    push(@HtmlStack, $code);
    $PageHtml .= "<$code>\n";
  }
  my $oldCode = pop(@HtmlStack);
  if ($oldCode ne $code) {
    $PageHtml .= "</$oldCode><$code>\n";
  }
  push(@HtmlStack, $code);
}

sub UnstackHtml {
  while (@HtmlStack > 0) {
    $PageHtml .=  "</" . pop(@HtmlStack) . ">\n";
  }
}

sub WikiToHTML {
  my ($pageText) = @_;
  my ($inPre);

  $inPre = 0;
  %SaveUrl = ();
  $SaveUrlIndex = 0;
  $PageHtml = "";
  $pageText =~ s/$FS//g;              # Remove separators (paranoia)
  if ($RawHtml) {
    $pageText =~ s/<html>((.|\n)*?)<\/html>/&StoreRaw($1)/ige;
  }
  $pageText = &QuoteHtml($pageText);
  # The <nowiki> tag stores text with no markup (except quoting HTML)
  $pageText =~
    s/\&lt;nowiki\&gt;((.|\n)*?)\&lt;\/nowiki\&gt;/&StoreRaw($1)/ige;
  $pageText =~ s/\\ *\r?\n/ /g;       # Join lines separated by backslash
  foreach (split(/\n/, $pageText)) {  # Process lines one-at-a-time
    $_ .= "\n";
    # Preformatted text handling
    if ($inPre) {
      if (/^\&lt;\/pre\&gt;\s*$/i) {  # </pre> end tag
        &UnstackHtml();
        $inPre = 0;
        next;
      }
      $PageHtml .= $_;
      next;                           # Skip any other markup
    }
    if (/^\&lt;pre\&gt;\s*$/i) {      # <pre> start tag
      $inPre = 1;
      &StackHtml("PRE", 1);
      next;
    }

    # Start-of-line markup patterns
    $SaveHtmlStack = 0;
    s/^\s*$/<p>\n/                && ($SaveHtmlStack = 1);
    s/^(\;+)([^:]+)\:/<dt>$2<dd>/ && &StackHtml("DL", length $1);
    s/^(\:+)/<dt><dd>/            && &StackHtml("DL", length $1);
    s/^(\*+)/<li>/                && &StackHtml("UL", length $1);
    s/^(\#+)/<li>/                && &StackHtml("OL", length $1);
    /^\s/                         && &StackHtml("PRE", 1);
    &UnstackHtml() if (!$SaveHtmlStack);

    # The strong markup pattern avoids overlapping tags with 5 quotes
    # by matching the inner quotes for the strong pattern.
    s/('*)'''(.*?)'''/$1<strong>$2<\/strong>/g;
    s/''(.*?)''/<em>$1<\/em>/g;
    s/\&lt;b\&gt;(.*?)\&lt;\/b\&gt;/<b>$1<\/b>/gi;
    s/\&lt;i\&gt;(.*?)\&lt;\/i\&gt;/<i>$1<\/i>/gi;
    s/----+/<hr>/g;
    s/$ImagePattern/&InlineImage($1)/ge;
    s/$UrlPattern/&StoreUrl($1)/ge;
    s/$InterLinkPattern/&InterPageLink($1)/geo;
    s/$LinkPattern/&GetPageOrEditLink($1)/geo;
    $PageHtml .= $_;
  }
  &UnstackHtml();
  $PageHtml =~ s/$FS(\d+)$FS/$SaveUrl{$1}/ge;
  return $PageHtml;
}

sub StoreRaw {
  my ($html) = @_;

  $SaveUrl{$SaveUrlIndex} = $html;
  return $FS . $SaveUrlIndex++ . $FS;
}

sub InlineImage {
  my ($name) = @_;

  $SaveUrl{$SaveUrlIndex} = "<img src=\"$name\">";
  return $FS . $SaveUrlIndex++ . $FS;
}

sub StoreUrl {
  my ($name) = @_;

  $SaveUrl{$SaveUrlIndex} = "<a href=\"$name\">$name</a>";
  return $FS . $SaveUrlIndex++ . $FS;
}

sub ReportError {
  my ($errmsg) = @_;

  print $q->header, "<H2>", $errmsg, "</H2>", $q->end_html;
}

sub ValidId {
  my ($id) = @_;

  if (!($id =~ /^$LinkPattern$/)) {
    ReportError("Invalid Page $id");
    return 0;
  }
  return 1;
}

sub OpenNewPage {
  my ($id) = @_;
  my $data;

  %page = ();
  $page{'text'} = "Describe the new page here.\n";
  $page{'timestamp'} = 943410730; # Must be fixed for edit conflicts.
  $page{'revision'} = 0;
  $FullPage{'current'} = join($FS3, %page);
}

sub OpenPage {
  my ($id) = @_;
  my ($fname, $data);

  if ($OpenId eq $id) {
    return;
  }
  $fname = &GetPageDirectory($id) . "$id.db";
  if (-f $fname) {
    open(IN, "<$fname") or die "Can't open $fname: $!";
    $data=<IN>;
    close IN;
    %FullPage = split(/$FS1/, $data, -1);  # -1 keeps trailing null fields
  } else {
    &OpenNewPage($id);
  }
  $OpenId = $id;
}

# Use this to get single-version pages.
sub GetPage {
  my ($id, $version) = @_;

  &OpenPage($id);
  return split(/$FS3/, $FullPage{$version}, -1);
}

sub GetCurrentPage {
  my ($id) = @_;

  return &GetPage($id,'current');
}

# Failing to get a cache lock is not a serious problem.
sub RequestCacheLock {
  my $n = 0;
  while (mkdir($LockDir, 0555) == 0) {
    $! == 17 || return 0; # EEXIST == 17 is OK, try later.
    $n++ < 4 || return 0;
    sleep(2);
  }
  return 1;
}

sub RequestLock {
  my $n = 0;
  while (mkdir($LockDir, 0555) == 0) {
    # EEXIST == 17 is OK, try later.
    $! ==  17 || die("can't make $LockDir: $!\n");
    $n++ < 10 || die("timed out waiting for $LockDir\n");
    sleep(3);
  }
}

sub ReleaseLock {
  rmdir($LockDir);
}

sub WriteStringToFile {
  my ($file, $string) = @_;

  open (OUT, ">$file") or die ("cant write $file: $!");
  print OUT  $string;
  close(OUT);
}

sub CreatePageDir {
  my ($dir, $id) = @_;
  my $subdir;

  $subdir = $dir . "/" . &GetPageDirectory($id);
  mkdir($subdir,0775)  if (!(-d $subdir));
  if ($id =~ m|([^/]+)/|) {
    $subdir = $subdir . "/" . $1;
    mkdir($subdir,0775)  if (!(-d $subdir));
  }
}

sub UpdateHtmlCache {
  my ($id, $html) = @_;
  my $id_file;

  $id_file = $HtmlDir . "/" . &GetPageDirectory($id) . "$id.htm";
  mkdir($HtmlDir,0775)  if (!(-d $HtmlDir));
  &CreatePageDir($HtmlDir, $id);
  if (&RequestCacheLock()) {
    &WriteStringToFile($id_file, $html);
    &ReleaseLock();
  }
}

sub CalcDay {
  my ($sec, $min, $hour, $mday, $mon, $year) = localtime(shift);

  return ("January", "February", "March", "April", "May", "June",
          "July", "August", "September", "October", "November",
          "December")[$mon]. " " . $mday . ", " . ($year+1900);
}

sub CalcDayNow {
  return CalcDay($^T);
}

sub CalcTime {
  my ($sec, $min, $hour, $mday, $mon, $year) = localtime(shift);
  my ($ampm, $mytz);

  $mytz = "";
  $mytz = " " . $ScriptTZ   if ($ScriptTZ ne "");
  $ampm = "am";
  if ($hour>11) {
    $ampm = "pm";
    $hour = $hour - 12;
  }
  $hour = 12   if ($hour == 0);
  $min = "0" . $min   if ($min<10);
  return $hour . ":" . $min . " " . $ampm . $mytz;
}

sub TimeToText {
  my ($t) = @_;

  return &CalcDay($t) . " " . &CalcTime($t);
}

sub ActionRc {
  my ($filedata, $rcline, $i, $daysago, $lastts, $ts);
  my @fullrc = ();
  my $starttime = 0;
  my $showbar = 0;

  if (!open(IN, "<$RcFile")) {
    print "<p><strong>Could not open RecentChanges log file: </strong>";
    print "$RcFile<p>Error was:\n<pre>$!</pre>\n";
    print "<p>Note: This error is normal if no changes have been made.\n";
    return;
  }
  $filedata=<IN>;
  close IN;
  @fullrc = split(/\n/, $filedata);
  ($lastts) = split(/$FS3/, $fullrc[$#fullrc]);
  $lastts++  if (($^T - $lastts) > 5);  # Skip last unless very recent

  if (&GetParam("from", 0)) {
    $starttime = &GetParam("from", 0);
    print "<h2>Updates since ", &TimeToText($starttime), "</h2>\n";
  }
  if (&GetParam("days", 0)) {
    $daysago = &GetParam("days", 0);
    $starttime = $^T - ((24*60*60)*$daysago);
    print "<h2>Updates in the last $daysago day",
          (($daysago != 1)?"s":""), "</h2>\n";
  }
  if ($starttime == 0) {
    $starttime = $^T - ((24*60*60)*$RcDefault);
    print "<h2>Updates in the last $RcDefault days</h2>\n";
  }

  foreach $i (@RcDays) {
    print " | "  if $showbar;
    $showbar = 1;
    print "<a href=\"$ScriptName\?action=rc&days=$i\">",
          $i, " day", (($i != 1)?"s":""), "</a>\n";
  }
  print "<br><a href=\"$ScriptName\?action=rc&from=$lastts\">",
        "List new changes starting from</a> ",
        &TimeToText($lastts), "<br>\n";

  for ($i = 0; $i < @fullrc ; $i++) {  # Later optimize loop?
    ($ts) = split(/$FS3/, $fullrc[$i]);
    last if ($ts >= $starttime);
  }
  if ($i == @fullrc) {
    print "<br><strong>No updates since ", &TimeToText($starttime),
          "</strong><br>\n";
  } else {
    splice(@fullrc, 0, $i);  # Remove items before index $i
    print &GetRcHtml(@fullrc);
  }
  print "<p>Page generated ", &TimeToText($^T), "<br>\n";
}

# Later update for kind and extended-data
sub GetRcHtml {
  my @outrc = @_;
  my ($rcline, $html, $date, $sum, $edit, $count, $newtop);
  my ($showedit, $inlist, $link, $all, $diffLink);
  my ($ts, $pagename, $summary, $isEdit, $host);
  my %changetime = ();
  my %pagecount = ();

  $showedit = &GetParam("showedit", $ShowEdits);
  if ($showedit != 1) {
    my @temprc = ();
    foreach $rcline (@outrc) {
      ($ts, $pagename, $summary, $isEdit, $host) = split(/$FS3/, $rcline);
      if ($showedit == 0) {  # 0 = No edits
        push(@temprc, $rcline)  if (!$isEdit);
      } else {               # 2 = Only edits
        push(@temprc, $rcline)  if ($isEdit);
      }
    }
    @outrc = @temprc;
  }

  foreach $rcline (@outrc) {
    ($ts, $pagename) = split(/$FS3/, $rcline);
    $pagecount{$pagename}++;
    $changetime{$pagename} = $ts;
  }
  $date = "";
  $inlist = 0;
  $html = "";
  $all = &GetParam("all", 0);
  $newtop = &GetParam("newtop", $RecentTop);
  $diffLink = &GetParam("difflink", 0);

  @outrc = reverse @outrc if ($newtop);
  foreach $rcline (@outrc) {
    ($ts, $pagename, $summary, $isEdit, $host) = split(/$FS3/, $rcline);
    next  if ((!$all) && ($ts < $changetime{$pagename}));
    if ($date ne &CalcDay($ts)) {
      $date = &CalcDay($ts);
      if ($inlist) {
        $html .= "</UL>\n";
        $inlist = 0;
      }
      $html .= "<p><strong>" . $date . "</strong><p>\n";
    }
    if (!$inlist) {
      $html .= "<UL>\n";
      $inlist = 1;
    }
    $host = &QuoteHtml($host);
    $sum = "";
    if (($summary ne "") && ($summary ne "*")) {
      $summary = &QuoteHtml($summary);
      $sum = "<strong>[$summary]</strong> "  if ($summary ne "*");
    }
    $edit = "";
    $edit = "<em>(edit)</em> "  if ($isEdit);
    $count = "";
    if ((!$all) && ($pagecount{$pagename} > 1)) {
      $count = "($pagecount{$pagename} changes) ";
    }
    if ($UseDiff) {
      $link = "<a href=\"$ScriptName\?action=browse&diff=1&"
              . "id=$pagename\">";
      if ($diffLink) {
        $link .= "$pagename</a>";
      } else {
        $link .= "(diff)</a>";
        $link .= "  <a href=\"$ScriptName\?$pagename\">$pagename</a>";
      }
    } else {
      $link = "<a href=\"$ScriptName\?$pagename\">$pagename</a>";
    }

    $html .= "<li>$link " .  &CalcTime($ts)
             . " $count$edit" . " $sum" . ". . . . . $host\n";
  }
  $html .= "</UL>\n" if ($inlist);
  return $html;
}

sub GetRedirectPage {
  my ($newid) = @_;
  my ($url, $html);

  $url = $FullUrl . "?" . $newid;
  $html = $q->redirect(-uri=>$url, -method=>"GET");
  $html .= "\nYour browser should go to the $newid page.";
  $html .= "  If it does not, click <a href=\"$url\">$newid</a>";
  $html .= " to continue.";
  return $html;
}

#END_OF_BROWSE_CODE

$OtherCode = ""; # Comment next line to always compile (slower)
#$OtherCode = <<'#END_OF_OTHER_CODE';
use Socket;  # For host lookups

sub DoOtherRequest {
  my ($id, $action, $text);

  $action = &GetParam("action", "");
  $id = &GetParam("id", "");
  if ($action ne "") {
    $action = lc($action);
    if      ($action eq "edit") {
      &ActionEdit($id, 0, 0, 0, "")  if &ValidId($id);
    } elsif ($action eq "editcopy") {
      &ActionEdit($id, 1, 0, 0, "")  if &ValidId($id);
    } elsif ($action eq "unlock") {
      &DoUnlock();
    } elsif ($action eq "index") {
      &DoIndex();
#   } elsif ($action eq "convert") {  # Normally not available
#     &DoConvert();
    } else {
      &ReportError("Invalid action parameter $action");
    }
    return;
  }
  $text = &GetParam("search", "");
  if ($text ne "") {
    &DoSearch($text);
    return;
  }
  # Handle posted pages
  $text = &GetParam("text", undef);
  $id = &GetParam("title", "");
  if ($text ne "") {
    &ActionPost($text)  if &ValidId($id);
    return;
  }
  &ReportError("Invalid URL.");
}

sub ActionEdit {
  my ($id, $editCopy, $isConflict, $oldTime, $newText) = @_;
  my $editRows = 20;
  my $oldText = "";
  my %pg = ();
  my $header;

  if (!$EditAllowed) {
    &ReportError("Editing not allowed: $SiteName is read-only.");
    return;
  }
  if ($editCopy) {
    %pg = &GetPage($id, 'authorcopy');
    $oldText = $pg{'text'};
    %pg = &GetCurrentPage($id); # Only the text is old
    $header = "Editing Old Copy of $id";
  } else {
    %pg = &GetCurrentPage($id);
    $oldText = $pg{'text'};
    $header = "Editing $id";
  }
  print &GetHeader("", &QuoteHtml($header), 0);
  if ($isConflict) {
    $editRows = 10;
    print "\n<H1>Edit Conflict!</H1>\n";
    if ($isConflict>1) {
      # The main purpose of a new warning is to display more text
      # and move the save button down from its old location.
      print "\n<H2>(This is a new conflict)</H2>\n";
    }
    print "<p><strong>Someone saved this page after you started editing.",
          " The top textbox contains the saved text.",
          " Only the text in the top textbox will be saved.</strong><br>\n",
          " Scroll down to see your edited version.<br>\n",
          "Last save time: ", &TimeToText($oldTime),
          " (Current time is: ", &TimeToText($^T), ")<br>\n";
  }
  print $q->startform("POST", "$ScriptName",
                      "application/x-www-form-urlencoded");
  $q->param("title", "$id");
  $q->param("oldtime", $pg{'timestamp'});
  $q->param("oldconflict", $isConflict);
  print $q->hidden("title"), "\n",
        $q->hidden("oldtime"), "\n",
        $q->hidden("oldconflict"), "\n",
        $q->textarea(-name=>'text', -default=>$oldText,
                     -rows=>$editRows, -columns=>65,
                     -override=>1, -wrap=>'virtual');
  print "<p>Summary:",
        $q->textfield(-name=>'summary',
                      -default=>'*', -override=>1,
                      -size=>60, -maxlength=>200);
  print "<br>", $q->checkbox(-name=>'recent_edit',
                            -label=>'This change is a minor edit.');
  print "<br>", $q->submit(-name=>'Save'), "\n";

  if ($isConflict) {
    print "\n<br><hr><p><strong>This is the text you submitted:",
          "</strong><p>",
          $q->textarea(-name=>'newtext', -default=>$newText,
                       -rows=>$editRows, -columns=>65,
                       -override=>1, -wrap=>'virtual'),
          "<p>\n";
  }
  print "<hr>\n";
  if ((!$editCopy)
       && defined($FullPage{'authorcopy'})
       && ($FullPage{'authorcopy'} ne "")) {
    print "<a href=\"$ScriptName?action=editcopy&id=$id\">",
          "Edit copy from previous author</a><br>\n";
  }
  print &GetGotoBar($id);
  print $q->endform;
  print $q->end_html;
}

sub DoIndex {
  print &GetHeader("","Index of all pages", 0);
  print "<br>";
  &PrintPageList(&AllPagesList());
  print "<hr>", &GetGotoBar("");
  print $q->end_html;
}

# XXX remove before release.
sub DoConvert {
  my ($name, @found, %pg);

  &RequestLock;
  print &GetHeader("","Converting all pages", 0);
  print "<br>";
  @found = ();
  foreach $name (&AllPagesList()) {
    %pg = &GetCurrentPage($name);
    push(@found, $name);
    $pg{text} =~ s/(^|\n)(\t+)\s+:\t/$1 . (":" x (length $2))/ge;
    $pg{text} =~
      s/(^|\n)(\t+)(.+):\t/$1 . (";" x (length $2)) . $3 . ":"/ge;
    $pg{text} =~ s/(^|\n)(\t+)\*/$1 . ("*" x (length $2))/ge;
    $pg{text} =~ s/(^|\n)(\t+)\d+\.?/$1 . ("#" x (length $2))/ge;
    $FullPage{'current'} = join($FS3, %pg);
    &SavePage($name);
  }
  &PrintPageList(@found);
  print "<hr>", &GetGotoBar("");
  print $q->end_html;
  &ReleaseLock;
}

sub DoSearch {
  my ($string) = @_;

  print &GetHeader("",&QuoteHtml("Search for: $string"), 0);
  print "<br>";
  &PrintPageList(&SearchTitleAndBody($string));
  print "<hr>", &GetGotoBar("");
  print $q->end_html;
}

sub PrintPageList {
  my $pagename;

  print "<h2>", ($#_ + 1), " pages found:</h2>\n";
  foreach $pagename (@_) {
    print ".... "  if ($pagename =~ m|/|);
    print &GetPageLink($pagename), "<br>\n";
  }
}

sub ActionPost {
  my ($string) = @_;

  my $id = &GetParam("title", "");
  my $summary = &GetParam("summary", "");
  my $oldtime = &GetParam("oldtime", "");
  my $oldconflict = &GetParam("oldconflict", "");
  my $is_edit = 0;
  my $edit_time = time;
  my $authorAddr = $ENV{REMOTE_ADDR};
  my ($edit_diff, $old, $newAuthor);

  if (!$EditAllowed) {
    &ReportError("Editing not allowed: $SiteName is read-only.");
    return;
  }
  %page = &GetCurrentPage($id);
  $old = $page{text};

  if ($old eq $string) {  # No changes (probably just pressed save)
    &ReBrowsePage($id, 0);
    return;
  }
  $string =~ s/$FS//;
  $summary =~ s/$FS//;
  $summary =~ s/[\r\n]//;
  $newAuthor = ($page{logaddr} ne $authorAddr);

  # Detect editing conflicts and resubmit edit
  if ($newAuthor && ($oldtime != $page{'timestamp'})) {
    if ($oldconflict>0) {  # Conflict again...
      &ActionEdit($id, 0, 2, $page{'timestamp'}, $string);
    } else {
      &ActionEdit($id, 0, 1, $page{'timestamp'}, $string);
    }
    return;
  }
  &RequestLock;

  $edit_diff = "";
  if ($UseDiff) {
    $edit_diff = &GetDiff($old, $string);
    &WriteDiff($id, $edit_time, $edit_diff);
  }
  &SetCurrentPage($id, $string, $newAuthor, $edit_diff);
  if (&GetParam("recent_edit", "") eq 'on') {
    $is_edit = 1;
  }
  &WriteRcLog($id, $summary, $is_edit, $edit_time);
  &ReleaseLock;
  &ReBrowsePage($id, 0);
}

sub ForceReleaseLock {
  my $n = 0;
  my $limit = 5;

  # First try to obtain lock (in case of normal edit lock)
  while (($n < $limit) && (mkdir($LockDir, 0555) == 0)) {
    $n++;
    sleep(3);
  }
  &ReleaseLock;  # Release the lock, even if we didn't get it.
  if ($n == $limit) {
    return 1;  # Lock was forced.
  }
  return 0;  # Lock was not forced.
}

# Note: call within lock.
sub SavePage {
  my ($id) = @_;
  my $file = &GetPageDirectory($id) . "$id.db";

  $FullPage{date} = &CalcDayNow;
  $FullPage{timestamp} = $^T;
  $FullPage{dbrev} = 1.1;
  $FullPage{pagerev}++;

  my $string = join($FS1, %FullPage);
  &CreatePageDir(".", $id);
  &WriteStringToFile($file, $string);
  if ($UseCache) {
    UnlinkHtmlCache($id);          # Old cached copy is invalid
    if ($FullPage{pagerev} == 1) { # If this is a new page...
      &NewPageCacheClear($id);     # ...uncache pages linked to this one.
    }
  }
}

sub SetCurrentPage {
  my ($id, $string, $newAuthor, $diff) = @_;

  %page = &GetCurrentPage($id);
  &SaveHistory($id, $newAuthor);
  $page{logaddr} = $ENV{REMOTE_ADDR};
  $page{logname} = &GetRemoteHost();
  $page{text} = $string;
  $page{date} = &CalcDayNow;
  $page{timestamp} = $^T;
  $page{prevdiff} = $diff;
  $page{revision}++;
  $FullPage{'current'} = join($FS3, %page);
  &SavePage($id);
}

sub SaveHistory {
  my ($id, $newAuthor) = @_;

  if ($page{revision} > 0) {
    if ($newAuthor) {
      $FullPage{'authorcopy'} = $FullPage{'current'};
    }
  }
}

sub GetRemoteHost {
  my ($rhost, $iaddr);

  $rhost = $ENV{REMOTE_HOST};
  if ($rhost eq "") {
    # Catch errors (including bad input) without aborting the script
    eval '$iaddr = inet_aton($ENV{REMOTE_ADDR});'
         . '$rhost = gethostbyaddr($iaddr, AF_INET)';
  }
  if ($rhost eq "") {
    $rhost = $ENV{REMOTE_ADDR};
    $rhost =~ s/\d+$/xxx/;      # Be somewhat anonymous
  }
  return $rhost;
}

sub AllPagesList {
  my (@pages, @dirs, $id, $dir);

  @dirs = qw(A B C D E F G H I J K L M N O P Q R S T U V W X Y Z other);
  foreach $dir (@dirs) {
    while (<$dir/*.db $dir/*/*.db>) {
      m|^[^/]+/(\S*).db|;
      $id = $1;
      push(@pages, $id);
    }
  }
  return sort(@pages);
}

sub SearchTitleAndBody {
  my ($string) = @_;
  my ($name, @found);

  foreach $name (&AllPagesList()) {
    my %pg = &GetCurrentPage($name);
    if (($pg{text} =~ /$string/i) || ($name =~ /$string/i)) {
      push(@found, $name);
    }
  }
  return @found;
}

sub SearchBody {
  my ($string) = @_;
  my ($name, @found);

  foreach $name (&AllPagesList()) {
    my %pg = &GetCurrentPage($name);
    if ($pg{text} =~ /$string/i){
      push(@found, $name);
    }
  }
  return @found;
}

sub UnlinkHtmlCache {
  my ($id) = @_;
  my $id_file;

  $id_file = $HtmlDir . "/" . &GetPageDirectory($id) . "$id.htm";
  if (-f $id_file) {
    unlink($id_file);
  }
}

sub NewPageCacheClear {
  my ($id) = @_;
  my $name;

  return if (!$UseCache);
  $id =~ s|.+/|/|;  # If subpage, search for just the subpage
  foreach $name (&SearchBody($id)) {
    &UnlinkHtmlCache($name);
  }
}

sub DoUnlock {
  my $LockMessage = "Normal Unlock.";

  print &GetHeader("","Removing edit lock", 0);
  print "<p>This operation may take several seconds...\n";
  if (&ForceReleaseLock()) {
    $LockMessage = "Forced Unlock.";
  }
  print "<br><h2>$LockMessage</h2><hr>", &GetGotoBar("");
  print $q->end_html;
}

# Note: all diff and recent-list operations should be done within locks.
sub WriteRcLog {
  my ($id, $summary, $is_edit, $edit_time) = @_;

  my $rhost = &GetRemoteHost();
  # The two fields at the end end of a line are kind and extension-hash
  my $rc_line = join ($FS3, $edit_time, $id, $summary,
                      $is_edit, $rhost, "0", "");
  if (!open (OUT, ">>$RcFile")) {
    die "RecentChanges log error: $!";
  }
  print OUT  $rc_line . "\n";
  close(OUT);
}

sub GetDiff {
  my ($old, $new) = @_;
  my $diff_out = "";

  &WriteStringToFile("old_diff", $old);
  &WriteStringToFile("new_diff", $new);

  $diff_out = `diff old_diff new_diff`;
  $diff_out =~ s/\\ No newline.*\n//g;   # Get rid of common complaint.
  # No need to unlink temp files--next diff will just overwrite.

  return $diff_out;
}

sub WriteDiff {
  my ($id, $edit_time, $diff_string) = @_;

  open (OUT, ">>diff_log") or die ("cant write diff_log");
  print OUT  "------\n" . $id . "|" . $edit_time . "\n";
  print OUT  $diff_string;
  close(OUT);
}
#END_OF_OTHER_CODE

# Common code (always compiled)
sub GetPageDirectory {
  my ($id) = @_;

  if ($id =~ /^([a-zA-Z])/) {
    return uc($1) . "/";
  }
  return "other/";
}

# Simple HTML cache
sub DoCacheBrowse {
  my ($query, $id_file, $text);

  return 0  if (!$UseCache);

  $query = $ENV{'QUERY_STRING'};
  if (($query eq "") && ($ENV{'REQUEST_METHOD'} eq "GET")) {
    $query = $HomePage;  # Allow caching of home page.
  }
  if (!($query =~ /^$LinkPattern$/)) {
    return 0;  # Only use cache for simple links
  }
  $id_file = $HtmlDir . "/" . &GetPageDirectory($query) . "$query.htm";
  if (-f "$id_file") {
    open(INFILE, "<$id_file") or return 0;
    $text = <INFILE>;
    close INFILE;
    print $text;
    return 1;
  }
  return 0;
}

if (!&DoCacheBrowse()) {
  eval $BrowseCode;
  if (!&DoBrowseRequest()) {
      eval $OtherCode;
      &DoOtherRequest();
  }
}
